/* opengl extensions */
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB = NULL;
PFNGLMULTITEXCOORD4FARBPROC glMultiTexCoord4fARB = NULL;

/* triangle data */
struct ogl_triangle_data {
	Bit32s smax, tmax;
	Bit32s ilod;
	Bit32u texbase;
	GLuint texID;
};

/* triangle vertex map data */
struct ogl_vertex_map_data {
	float s,t,w;
};

/* triangle vertex data */
struct ogl_vertex_data {
	float r,g,b,a,z,w,d;
	ogl_vertex_map_data m[2];
};

/* texture cache buffer */
Bit32u texrgb[256*256];

/* texture address mapping */
std::map <const Bit32u, GLuint> textures;



#define get_depth(VV, FBZCOLORPATH, FBZMODE, ITERZ, ITERW, depthval)			\
	do {																		\
	Bit32s wfloat;																\
	/* compute "floating point" W value (used for depth and fog) */				\
	if ((ITERW) & U64(0xffff00000000))											\
		wfloat = 0x0000;														\
	else																		\
	{																			\
		Bit32u temp = (Bit32u)(ITERW);											\
		if ((temp & 0xffff0000) == 0)											\
			wfloat = 0xffff;													\
		else																	\
		{																		\
			int exp = count_leading_zeros(temp);								\
			wfloat = ((exp << 12) | ((~temp >> (19 - exp)) & 0xfff)) + 1;		\
		}																		\
	}																			\
																				\
	/* compute depth value (W or Z) for this pixel */							\
	if (FBZMODE_WBUFFER_SELECT(FBZMODE) == 0)									\
		CLAMPED_Z(ITERZ, FBZCOLORPATH, depthval);								\
	else if (FBZMODE_DEPTH_FLOAT_SELECT(FBZMODE) == 0)							\
		depthval = wfloat;														\
	else																		\
	{																			\
		if ((ITERZ) & 0xf0000000)												\
			depthval = 0x0000;													\
		else																	\
		{																		\
			Bit32u temp = (ITERZ) << 4;											\
			if ((temp & 0xffff0000) == 0)										\
				depthval = 0xffff;												\
			else																\
			{																	\
				int exp = count_leading_zeros(temp);							\
				depthval = ((exp << 12) | ((~temp >> (19 - exp)) & 0xfff)) + 1;	\
			}																	\
		}																		\
	}																			\
																				\
	/* add the bias */															\
	if (FBZMODE_ENABLE_DEPTH_BIAS(FBZMODE))										\
	{																			\
		depthval += (Bit16s)(VV)->reg[zaColor].u;								\
		CLAMP(depthval, 0, 0xffff);												\
	}																			\
	if (FBZMODE_DEPTH_SOURCE_COMPARE(FBZMODE))									\
		depthval = (Bit16u)(VV)->reg[zaColor].u;								\
	}while(0);


void get_lod(const void *extradata, ogl_triangle_data *td, const int i) {
	const poly_extra_data *extra = (const poly_extra_data *)extradata;
	voodoo_state *v=(voodoo_state*)extra->state;
	Bit32s oow, lod, ilod;
	Bit32s smax, tmax;
	Bit32u texmode=v->tmu[i].reg[textureMode].u;
	Bit32s lodbase= i==0?extra->lodbase0:extra->lodbase1;
	tmu_state *tt=&extra->state->tmu[i];
	Bit64s iterw;

	iterw=i==0?extra->startw0:extra->startw1;

	/* determine the S/T/LOD values for this texture */
	if (TEXMODE_ENABLE_PERSPECTIVE(texmode))
	{
		oow = fast_reciplog((iterw), &lod);
		lod += (lodbase);
	}
	else
	{
		lod = (lodbase);
	}

	/* clamp the LOD */
	lod += tt->lodbias;
	if (lod < tt->lodmin)
		lod = tt->lodmin;
	if (lod > tt->lodmax)
		lod = tt->lodmax;

	/* now the LOD is in range; if we don't own this LOD, take the next one */
	ilod = lod >> 8;
	if (!((tt->lodmask >> ilod) & 1))
		ilod++;

	/* fetch the texture base */
	td->texbase = tt->lodoffset[ilod];

	/* compute the maximum s and t values at this LOD */
	td->smax = (tt->wmask >> ilod) + 1;
	td->tmax = (tt->hmask >> ilod) + 1;

	td->ilod = ilod;
	
}


void get_vertex_data(Bit32u x, Bit32u y, const void *extradata, ogl_triangle_data *td, ogl_vertex_data *vd) {
	const poly_extra_data *extra = (const poly_extra_data *)extradata;
	Bit32s iterr, iterg, iterb, itera;
	Bit32s iterz, d;
	Bit64s iterw;
	Bit32s dx, dy;
	Bit32s scry;

	dx = x - extra->ax;
	dy = y - extra->ay;
	iterr = extra->startr + ((dy * extra->drdy)>>4) + ((dx * extra->drdx)>>4);
	iterg = extra->startg + ((dy * extra->dgdy)>>4) + ((dx * extra->dgdx)>>4);
	iterb = extra->startb + ((dy * extra->dbdy)>>4) + ((dx * extra->dbdx)>>4);
	itera = extra->starta + ((dy * extra->dady)>>4) + ((dx * extra->dadx)>>4);
	iterz = extra->startz + ((dy * extra->dzdy)>>4) + ((dx * extra->dzdx)>>4);
	iterw = extra->startw + ((dy * extra->dwdy)>>4) + ((dx * extra->dwdx)>>4);

	for (int i=0;i<extra->texcount;i++) {
		voodoo_state *v=(voodoo_state*)extra->state;
		Bit64s iters,itert,iterw;
		Bit32s s, t, lod;

		Bit32u texmode=v->tmu[i].reg[textureMode].u;
		Bit32s lodbase= i==0?extra->lodbase0:extra->lodbase1;

		Bit32s oow;
		Bit32s ilod = td[i].ilod;

		iterw = v->tmu[i].startw + ((dy * v->tmu[i].dwdy)>>4) + ((dx * v->tmu[i].dwdx)>>4);
		iters = v->tmu[i].starts + ((dy * v->tmu[i].dsdy)>>4) + ((dx * v->tmu[i].dsdx)>>4);
		itert = v->tmu[i].startt + ((dy * v->tmu[i].dtdy)>>4) + ((dx * v->tmu[i].dtdx)>>4);

		/* determine the S/T/LOD values for this texture */
		if (TEXMODE_ENABLE_PERSPECTIVE(texmode))
		{
			oow = fast_reciplog((iterw), &lod);
			s = ((Bit64s)oow * (iters)) >> 29;
			t = ((Bit64s)oow * (itert)) >> 29;
		}
		else
		{
			s = (iters) >> 14;
			t = (itert) >> 14;
		}

		/* clamp W */
		if (TEXMODE_CLAMP_NEG_W(texmode) && (iterw) < 0)
			s = t = 0;

		vd->m[i].s = (float)((float)s/(float)(td[i].smax*(1<<(18+ilod))));
		vd->m[i].t = (float)((float)t/(float)(td[i].tmax*(1<<(18+ilod))));
		vd->m[i].w = (float)((float)iterw/(float)(0xffffff));
	};

	//if (FBZCP_CC_LOCALSELECT(v->reg[fbzColorPath].u)) {
	//	vd->r = (float)v->reg[color0].rgb.r / (float)0xff;
	//	vd->g = (float)v->reg[color0].rgb.g / (float)0xff;
	//	vd->b = (float)v->reg[color0].rgb.b / (float)0xff;
	//} else 
	{
	vd->r = (float)((float)iterr/(float)(1<<20));
	vd->g = (float)((float)iterg/(float)(1<<20));
	vd->b = (float)((float)iterb/(float)(1<<20));
	}
	vd->a = (float)((float)itera/(float)(1<<20));
	vd->z = (float)((float)iterz/(float)(1<<20));
	vd->w = (float)((float)iterw/(float)(0xffffff));

	get_depth(v, v->reg[fbzColorPath].u, v->reg[fbzMode].u, iterz, iterw, d)
	vd->d = (float)((float)d/(float)(0xffff));

}

void ogl_cache_texture(const void *extradata, ogl_triangle_data *td) {
	const poly_extra_data *extra = (const poly_extra_data *)extradata;
	voodoo_state *v=(voodoo_state*)extra->state;
	Bit32u texbase;	

	Bit32s smax, tmax;
	GLuint texture;
	Bit8u *texptr8;
	Bit16u *texptr16;
	Bit32u *texrgbp;
	GLuint texID;

	Bit32u TEXMODE;

	for (int j=0;j<extra->texcount;j++) {
		TEXMODE = j==0 ? extra->r_textureMode0 : extra->r_textureMode1;
		texbase = td[j].texbase;
		smax=td[j].smax;
		tmax=td[j].tmax;

		if ( v->tmu[j].lodmin < (8 << 8) )
		if (textures.find(texbase) == textures.end()) {
			texptr8 = (Bit8u *)&v->tmu[j].ram[(texbase) & v->tmu[j].mask];
			texptr16 = (Bit16u *)&v->tmu[j].ram[(texbase) & v->tmu[j].mask];
			texrgbp = (Bit32u *)&texrgb[0];
			memset(texrgbp,0,256*256*4);
			if (TEXMODE_FORMAT(v->tmu[j].reg[textureMode].u) < 8)
				for (int i=0; i<(smax*tmax); i++) {
					Bitu data = v->tmu[j].lookup[*texptr8++];
					if (!(FBZMODE_ENABLE_CHROMAKEY(v->reg[fbzMode].u) && ((data & 0x00ffffff) == (v->reg[chromaKey].u & 0x00ffffff))))
						*texrgbp = (data&0xff000000) | ((data<<16)&0x00ff0000) | (data&0x0000ff00) | ((data>>16)&0x000000ff);
					texrgbp++;
				}
			else if (TEXMODE_FORMAT(v->tmu[j].reg[textureMode].u) >= 10 && TEXMODE_FORMAT(v->tmu[j].reg[textureMode].u) <= 12)
				for (int i=0; i<(smax*tmax); i++) {
					Bitu data = v->tmu[j].lookup[*texptr16++];
					if (!(FBZMODE_ENABLE_CHROMAKEY(v->reg[fbzMode].u) && ((data & 0x00ffffff) == (v->reg[chromaKey].u & 0x00ffffff))))
						*texrgbp = (data&0xff000000) | ((data<<16)&0x00ff0000) | (data&0x0000ff00) | ((data>>16)&0x000000ff);
					texrgbp++;
				}
			else {
				for (int i=0; i<(smax*tmax); i++) {
					Bitu data = (v->tmu[j].lookup[*texptr16 & 0xFF] & 0xFFFFFF) | ((*texptr16 & 0xff00) << 16);
					if (!(FBZMODE_ENABLE_CHROMAKEY(v->reg[fbzMode].u) && ((data & 0x00ffffff) == (v->reg[chromaKey].u & 0x00ffffff))))
						*texrgbp = (data&0xff000000) | ((data<<16)&0x00ff0000) | (data&0x0000ff00) | ((data>>16)&0x000000ff);
					texrgbp++;
					texptr16++;
				}
			}


			texrgbp = (Bit32u *)&texrgb[0];
			texID = textures.size() + 1;

			glGenTextures(1, &texID);
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glBindTexture(GL_TEXTURE_2D, texID);
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST+TEXMODE_MINIFICATION_FILTER(TEXMODE)); 
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST+TEXMODE_MAGNIFICATION_FILTER(TEXMODE)); 
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT/*-TEXMODE_CLAMP_S(TEXMODE)*/); 
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT/*-TEXMODE_CLAMP_T(TEXMODE)*/);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, smax, tmax, 0, GL_RGBA, GL_UNSIGNED_BYTE, texrgbp);

			textures[texbase] = texID;

		} else {
			texID = textures[texbase];
		}

		td[j].texID = texID;
	}
}



void ogl_draw_triangle(poly_extra_data *extra) {
	voodoo_state *v=extra->state;
	ogl_triangle_data td[2];
	ogl_vertex_data vd[3];

	for (char i=0;i<extra->texcount;i++)
		get_lod((void*)extra,&td[i],i);

	if (extra->texcount > 0) {
		
		ogl_cache_texture((void*)extra,td);

		glActiveTextureARB(GL_TEXTURE0_ARB);
		glBindTexture (GL_TEXTURE_2D, td[0].texID);
		glEnable (GL_TEXTURE_2D);

		if ( v->tmu[1].lodmin < (8 << 8) ) {
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glBindTexture (GL_TEXTURE_2D, td[1].texID);
		glEnable (GL_TEXTURE_2D);
		}

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);

	}

	else {
		glDisable (GL_TEXTURE_2D);
		glDisable(GL_BLEND);
	}

	glBegin(GL_TRIANGLES);

	get_vertex_data(v->fbi.ax, v->fbi.ay, (void*)extra, td, &vd[0]);
	get_vertex_data(v->fbi.bx, v->fbi.by, (void*)extra, td, &vd[1]);
	get_vertex_data(v->fbi.cx, v->fbi.cy, (void*)extra, td, &vd[2]);

	glColor3f(vd[0].r,vd[0].g,vd[0].b); 
	glMultiTexCoord4fARB(GL_TEXTURE0_ARB,vd[0].m[0].s*vd[0].m[0].w, vd[0].m[0].t*vd[0].m[0].w, 0.0f, vd[0].m[0].w); 
	if ( v->tmu[1].lodmin < (8 << 8) ) 
		glMultiTexCoord4fARB(GL_TEXTURE1_ARB,vd[0].m[1].s*vd[0].m[1].w, vd[0].m[1].t*vd[0].m[1].w, 0.0f, vd[0].m[1].w); 
	glVertex3f((float)v->fbi.ax/16.0f, (float)v->fbi.ay/16.0f, vd[0].d); 

	glColor3f(vd[1].r,vd[1].g,vd[1].b); 
	glMultiTexCoord4fARB(GL_TEXTURE0_ARB,vd[1].m[0].s*vd[1].m[0].w, vd[1].m[0].t*vd[1].m[0].w, 0.0f, vd[1].m[0].w); 
	if ( v->tmu[1].lodmin < (8 << 8) )
		glMultiTexCoord4fARB(GL_TEXTURE1_ARB,vd[1].m[1].s*vd[1].m[1].w, vd[1].m[1].t*vd[1].m[1].w, 0.0f, vd[1].m[1].w); 
	glVertex3f((float)v->fbi.bx/16.0f, (float)v->fbi.by/16.0f, vd[1].d); 

	glColor3f(vd[2].r,vd[2].g,vd[2].b); 
	glMultiTexCoord4fARB(GL_TEXTURE0_ARB,vd[2].m[0].s*vd[2].m[0].w, vd[2].m[0].t*vd[2].m[0].w, 0.0f, vd[2].m[0].w); 
	if ( v->tmu[1].lodmin < (8 << 8) )
		glMultiTexCoord4fARB(GL_TEXTURE1_ARB,vd[2].m[1].s*vd[2].m[1].w, vd[2].m[1].t*vd[2].m[1].w, 0.0f, vd[2].m[1].w); 
	glVertex3f((float)v->fbi.cx/16.0f, (float)v->fbi.cy/16.0f, vd[2].d); 

	
	glEnd();
	glDisable (GL_TEXTURE_2D);

}


void ogl_swap_buffer() {
	SDL_GL_SwapBuffers ( );
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glClearDepth(1.0f);
}


void ogl_texture_clear(Bit32u texbase) {
//	if (!textures.empty()) textures.clear();
	std::map<const Bit32u, GLuint>::iterator t;
	t=textures.find(texbase);
	if (t != textures.end()) {
		glDeleteTextures(1,&t->second);
		textures.erase(t);
	}

}

void ogl_draw_pixel (int x, int y, int r, int g, int b) {
	glColor3f((float)r/256, (float)g/256, (float)b/256);
	glBegin(GL_POINTS);
	glVertex2i(x, y);
	glEnd();
}

void ogl_set_window(voodoo_state *v) {

	if (FBZMODE_ENABLE_DEPTHBUF(v->reg[fbzMode].u)) {
		glClearDepth(1.0f);									// Depth Buffer Setup
		glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
//		glDepthFunc(GL_LEQUAL);								// The Type Of Depth Testing To Do
		glDepthFunc(GL_NEVER+FBZMODE_DEPTH_FUNCTION(v->reg[fbzMode].u));// The Type Of Depth Testing To Do
	}
	else
		glDisable(GL_DEPTH_TEST);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations

//glHint(GL_CLIP_VOLUME_CLIPPING_HINT_EXT,GL_FASTEST);

	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity( );
	if (FBZMODE_Y_ORIGIN(v->reg[fbzMode].u))
		glOrtho( 0, v->fbi.width, 0, v->fbi.height, 0.0f, -1.0f );
	else
		glOrtho( 0, v->fbi.width, v->fbi.height, 0, 0.0f, -1.0f );
	glViewport( 0, 0, v->fbi.width, v->fbi.height );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity( );
}

void ogl_init(voodoo_state *v) {
	std::map<const Bit32u, GLuint>::iterator t;
//SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	if (SDL_SetVideoMode(v->fbi.width, v->fbi.height, 32, SDL_OPENGL) == 0)
		E_Exit("opengl init error");

	for (t=textures.begin(); t!=textures.end(); t++)
		glDeleteTextures(1,&t->second);
	if (!textures.empty()) textures.clear();

	ogl_set_window(v);

// TODO check for extension support
	if (!glActiveTextureARB)
	glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");

	if (!glMultiTexCoord4fARB)
	glMultiTexCoord4fARB = (PFNGLMULTITEXCOORD4FARBPROC)wglGetProcAddress("glMultiTexCoord4fARB");


}